/**
 * vim: set ts=4 :
 * =============================================================================
 * SourceMod SDKTools Extension
 * Copyright (C) 2004-2010 AlliedModders LLC.  All rights reserved.
 * =============================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * As a special exception, AlliedModders LLC gives you permission to link the
 * code of this program (as well as its derivative works) to "Half-Life 2," the
 * "Source Engine," the "SourcePawn JIT," and any Game MODs that run on software
 * by the Valve Corporation.  You must obey the GNU General Public License in
 * all respects for all other code used.  Additionally, AlliedModders LLC grants
 * this exception to all derivative works.  AlliedModders LLC defines further
 * exceptions, found in LICENSE.txt (as of this writing, version JULY-31-2007),
 * or <http://www.sourcemod.net/license.php>.
 *
 * Version: $Id$
 */

#include "extension.h"
#include "rulesnatives.h"

const char *g_szGameRulesProxy;

const char *GetDTTypeName(int type)
{
	switch (type)
	{
	case DPT_Int:
		{
			return "integer";
		}
	case DPT_Float:
		{
			return "float";
		}
	case DPT_Vector:
		{
			return "vector";
		}
	case DPT_String:
		{
			return "string";
		}
	case DPT_Array:
		{
			return "array";
		}
	case DPT_DataTable:
		{
			return "datatable";
		}
	default:
		{
			return NULL;
		}
	}

	return NULL;
}

void GameRulesInit()
{
	g_szGameRulesProxy = g_pGameConf->GetKeyValue("GameRulesProxy");
}

enum PropFieldType
{
	PropField_Unsupported,		/**< The type is unsupported. */
	PropField_Integer,			/**< Valid for SendProp and Data fields */
	PropField_Float,			/**< Valid for SendProp and Data fields */
	PropField_Entity,			/**< Valid for Data fields only (SendProp shows as int) */
	PropField_Vector,			/**< Valid for SendProp and Data fields */
	PropField_String,			/**< Valid for SendProp and Data fields */
	PropField_String_T,			/**< Valid for Data fields.  Read only! */
};

static cell_t GameRules_GetProp(IPluginContext *pContext, const cell_t *params)
{
	char *prop;
	int offset;
	int bit_count;
	int element = params[3];

	void *pGameRules = g_pSDKTools->GetGameRules();
	if (!pGameRules || !g_szGameRulesProxy || !strcmp(g_szGameRulesProxy, ""))
		return pContext->ThrowNativeError("Gamerules lookup failed.");

	pContext->LocalToString(params[1], &prop);

	sm_sendprop_info_t info;
	
	gamehelpers->FindSendPropInfo(g_szGameRulesProxy, prop, &info);

	offset = info.actual_offset;
	int elementCount = 1;
	SendTable *pTable;
	SendProp *pProp;
	bit_count = info.prop->m_nBits;

	switch (info.prop->GetType())
	{
	case DPT_Int:
		if (element > 0)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);
		break;
	case DPT_DataTable:
		pTable = info.prop->GetDataTable();
		if (!pTable)
			return pContext->ThrowNativeError("Error looking up DataTable.");
		
		elementCount = pTable->GetNumProps();
		if (element >= elementCount)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);

		pProp = pTable->GetProp(element);
		if (pProp->GetType() != DPT_Int)
			return pContext->ThrowNativeError("Prop %s is not an integer (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Int);

		offset += pProp->GetOffset();
		bit_count = pProp->m_nBits;
		break;
	default:
		return pContext->ThrowNativeError("Prop %s is not an integer (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Int);
	}

	if (bit_count < 1)
	{
		bit_count = params[3] * 8;
	}

	if (bit_count >= 17)
	{
		return *(int32_t *)((intptr_t)pGameRules + offset);
	}
	else if (bit_count >= 9)
	{
		return *(int16_t *)((intptr_t)pGameRules + offset);
	}
	else if (bit_count >= 2)
	{
		return *(int8_t *)((intptr_t)pGameRules + offset);
	}
	else
	{
		return *(bool *)((intptr_t)pGameRules + offset) ? 1 : 0;
	}

	return -1;
}

static cell_t GameRules_GetPropFloat(IPluginContext *pContext, const cell_t *params)
{
	char *prop;
	int offset;
	int element = params[2];

	void *pGameRules = g_pSDKTools->GetGameRules();
	if (!pGameRules || !g_szGameRulesProxy || !strcmp(g_szGameRulesProxy, ""))
		return pContext->ThrowNativeError("Gamerules lookup failed.");

	pContext->LocalToString(params[1], &prop);

	sm_sendprop_info_t info;

	gamehelpers->FindSendPropInfo(g_szGameRulesProxy, prop, &info);

	offset = info.actual_offset;
	int elementCount = 1;
	SendTable *pTable;
	SendProp *pProp;

	switch (info.prop->GetType())
	{
	case DPT_Float:
		if (element > 0)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);
		break;
	case DPT_DataTable:
		pTable = info.prop->GetDataTable();
		if (!pTable)
			return pContext->ThrowNativeError("Error looking up DataTable.");
		
		elementCount = pTable->GetNumProps();
		if (element >= elementCount)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);

		pProp = pTable->GetProp(element);
		if (pProp->GetType() != DPT_Float)
			return pContext->ThrowNativeError("Prop %s is not a float (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Float);

		offset += pProp->GetOffset();
		break;
	default:
		return pContext->ThrowNativeError("Prop %s is not a float (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Float);
	}

	float val = *(float *)((intptr_t)pGameRules + offset);

	return sp_ftoc(val);
}

static cell_t GameRules_SetProp(IPluginContext *pContext, const cell_t *params)
{
	char *prop;
	int offset;
	int bit_count;
	int element = params[4];

	bool sendChange = true;
	if (params[5] == 0)
		sendChange = false;

	CBaseEntity *pProxy = NULL;
	if (sendChange && ((pProxy = GetProxyEnt()) == NULL))
		return pContext->ThrowNativeError("Couldn't find gamerules proxy entity");
	
	void *pGameRules = g_pSDKTools->GetGameRules();
	if (!pGameRules || !g_szGameRulesProxy || !strcmp(g_szGameRulesProxy, ""))
		return pContext->ThrowNativeError("Gamerules lookup failed");

	pContext->LocalToString(params[1], &prop);

	sm_sendprop_info_t info;
	
	gamehelpers->FindSendPropInfo(g_szGameRulesProxy, prop, &info);

	offset = info.actual_offset;
	int elementCount = 1;
	SendTable *pTable;
	SendProp *pProp;
	bit_count = info.prop->m_nBits;

	switch (info.prop->GetType())
	{
	case DPT_Int:
		if (element > 0)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);
		break;
	case DPT_DataTable:
		pTable = info.prop->GetDataTable();
		if (!pTable)
			return pContext->ThrowNativeError("Error looking up DataTable.");
		
		elementCount = pTable->GetNumProps();
		if (element >= elementCount)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);

		pProp = pTable->GetProp(element);
		if (pProp->GetType() != DPT_Int)
			return pContext->ThrowNativeError("Prop %s is not an integer (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Int);

		offset += pProp->GetOffset();
		bit_count = pProp->m_nBits;
		break;
	default:
		return pContext->ThrowNativeError("Prop %s is not an integer (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Int);
	}

	if (bit_count < 1)
	{
		bit_count = params[3] * 8;
	}

	if (bit_count >= 17)
	{
		*(int32_t *)((intptr_t)pGameRules + offset) = params[2];
		if (sendChange)
		{
			*(int32_t *)((intptr_t)pProxy + offset) = params[2];
			gamehelpers->SetEdictStateChanged(gamehelpers->EdictOfIndex(gamehelpers->EntityToBCompatRef(pProxy)), offset);
		}
	}
	else if (bit_count >= 9)
	{
		*(int16_t *)((intptr_t)pGameRules + offset) = (int16_t)params[2];
		if (sendChange)
		{
			*(int16_t *)((intptr_t)pProxy + offset) = (int16_t)params[2];
			gamehelpers->SetEdictStateChanged(gamehelpers->EdictOfIndex(gamehelpers->EntityToBCompatRef(pProxy)), offset);
		}
	}
	else if (bit_count >= 2)
	{
		*(int8_t *)((intptr_t)pGameRules + offset) = (int8_t)params[2];
		if (sendChange)
		{
			*(int8_t *)((intptr_t)pProxy + offset) = (int8_t)params[2];
			gamehelpers->SetEdictStateChanged(gamehelpers->EdictOfIndex(gamehelpers->EntityToBCompatRef(pProxy)), offset);
		}
	}
	else
	{
		*(bool *)((intptr_t)pGameRules + offset) = (params[2] == 0) ? false : true;
		if (sendChange)
		{
			*(bool *)((intptr_t)pProxy + offset) = (params[2] == 0) ? false : true;
			gamehelpers->SetEdictStateChanged(gamehelpers->EdictOfIndex(gamehelpers->EntityToBCompatRef(pProxy)), offset);
		}
	}

	return 0;
}

static cell_t GameRules_SetPropFloat(IPluginContext *pContext, const cell_t *params)
{
	char *prop;
	int offset;
	int element = params[3];

	bool sendChange = true;
	if (params[4] == 0)
		sendChange = false;

	CBaseEntity *pProxy = NULL;
	if (sendChange && ((pProxy = GetProxyEnt()) == NULL))
		return pContext->ThrowNativeError("Couldn't find gamerules proxy entity.");
	
	void *pGameRules = g_pSDKTools->GetGameRules();
	if (!pGameRules || !g_szGameRulesProxy || !strcmp(g_szGameRulesProxy, ""))
		return pContext->ThrowNativeError("Gamerules lookup failed.");

	pContext->LocalToString(params[1], &prop);

	sm_sendprop_info_t info;

	gamehelpers->FindSendPropInfo(g_szGameRulesProxy, prop, &info);

	offset = info.actual_offset;
	int elementCount = 1;
	SendTable *pTable;
	SendProp *pProp;

	switch (info.prop->GetType())
	{
	case DPT_Float:
		if (element > 0)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);
		break;
	case DPT_DataTable:
		pTable = info.prop->GetDataTable();
		if (!pTable)
			return pContext->ThrowNativeError("Error looking up DataTable.");
		
		elementCount = pTable->GetNumProps();
		if (element >= elementCount)
			return pContext->ThrowNativeError("Element %d is out of bounds.", element);

		pProp = pTable->GetProp(element);
		if (pProp->GetType() != DPT_Float)
			return pContext->ThrowNativeError("Prop %s is not a float! (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Float);

		offset += pProp->GetOffset();
		break;
	default:
		return pContext->ThrowNativeError("Prop %s is not a float (%d != %d)",
			prop,
			info.prop->GetType(),
			DPT_Float);
	}

	float newVal = sp_ctof(params[2]);

	*(float *)((intptr_t)pGameRules + offset) = newVal;

	if (sendChange)
	{
		*(float *)((intptr_t)pProxy + offset) = newVal;
		gamehelpers->SetEdictStateChanged(gamehelpers->EdictOfIndex(gamehelpers->EntityToBCompatRef(pProxy)), offset);
	}

	return 0;
}

static CBaseEntity *FindEntityByNetClass(int start, const char *classname)
{
	// TODO: get maxents from gpGlobals instead of hardcoding
	for (int i = start; i < 2048; i++)
	{
		edict_t *current = gamehelpers->EdictOfIndex(i);
		if (current == NULL)
			continue;

		IServerNetworkable *network = current->GetNetworkable();
		if (network == NULL)
			continue;

		ServerClass *sClass = network->GetServerClass();
		const char *name = sClass->GetName();

		if (!strcmp(name, classname))
			return gamehelpers->ReferenceToEntity(gamehelpers->IndexOfEdict(current));		
	}

	return NULL;
}

static CBaseEntity* GetProxyEnt()
{
	static cell_t proxyEntRef = -1;
	CBaseEntity *pProxy;
	if (proxyEntRef == -1 || (pProxy = gamehelpers->ReferenceToEntity(proxyEntRef)) == NULL)
	{
		pProxy = FindEntityByNetClass(playerhelpers->GetMaxClients(), g_szGameRulesProxy);
		proxyEntRef = gamehelpers->EntityToReference(pProxy);
	}
	
	return pProxy;
}


sp_nativeinfo_t g_GameRulesNatives[] = 
{
	{"GameRules_GetProp",		GameRules_GetProp},
	{"GameRules_GetPropFloat",	GameRules_GetPropFloat},
	{"GameRules_SetProp",		GameRules_SetProp},
	{"GameRules_SetPropFloat",	GameRules_SetPropFloat},
	{NULL,						NULL},
};

